import os
from pytnm.utils import shapefile2
# from arcpy import CreateFeatureclass_management as cf
# from arcpy import AddField_management as af
# from arcpy import env

# env.overwriteOutput = True

# https://www.esri.com/library/whitepapers/pdfs/shapefile2.pdf
# NULL = 0
# POINT = 1
# POLYLINE = 3
# POLYGON = 5
# MULTIPOINT = 8
# POINTZ = 11
# POLYLINEZ = 13
# POLYGONZ = 15
# MULTIPOINTZ = 18
# POINTM = 21
# POLYLINEM = 23
# POLYGONM = 25
# MULTIPOINTM = 28
# MULTIPATCH = 31

def create_receivers(input_folder):
    f = "receiver"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POINTZ)
    writer.field("rec_id", "C", size=10)
    writer.field("x", "N")
    writer.field("y", "N")
    writer.field("z", "N")
    writer.field("du", "N")
    writer.field("bldg_hgt", "N")
    writer.field("land_use", "C", size=100)
    writer.field("nac_cat", "C", size=1)
    writer.field("nac_lvl", "N")
    writer.field("ext_use", "C", size=1)
    writer.field("displace", "C", size=1)
    writer.field("bldg_row", "N")
    writer.field("exist_snd", "N")
    writer.field("nobld_snd", "N")
    writer.field("bld_snd", "N")
    writer.field("ex_imp", "C", size=1)
    writer.field("nobld_imp", "C", size=1)
    writer.field("bld_imp", "C", size=1)
    writer.field("bar_name", "C", size=10)
    writer.field("bar_reduct", "N")
    writer.field("project", "C", size=25)
    return f"{fn}.shp"

def create_existing_roadway(input_folder):
    f = "existing_roadway"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POLYLINEZ)
    writer.field("road_name", "C", size=32)
    writer.field("exist_traf", "N")
    writer.field("nobld_traf", "N")
    writer.field("div_lanes", "N")
    writer.field("speed", "N")
    writer.field("ex_total", "N")
    writer.field("nb_total", "N")
    writer.field("auto_ex", "N")
    writer.field("med_ex", "N")
    writer.field("heavy_ex", "N")
    writer.field("auto_nobld", "N")
    writer.field("med_nobld", "N")
    writer.field("heavy_nobld", "N")
    writer.field("heavy_pct", "N")
    writer.field("medium_pct", "N")
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_proposed_roadway(input_folder):
    f = "proposed_roadways"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POLYLINEZ)
    writer.field("road_name", "C", size=32)
    writer.field("bld_traf", "N")
    writer.field("div_lanes", "N")
    writer.field("speed", "N")
    writer.field("bld_total", "N")
    writer.field("auto_bld", "N")
    writer.field("med_bld", "N")
    writer.field("heavy_bld", "N")
    writer.field("heavy_pct", "N")
    writer.field("medium_pct", "N")
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_barrier(input_folder):
    f = "barrier"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POLYLINEZ)
    writer.field("name", "C", size=32)
    writer.field("feasible", "C", size=1)
    writer.field("reasonable", "C", size=1)
    writer.field("pert_inc", "N")
    writer.field("pert_num", "N")
    writer.field("init_hgt", "N")
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_terrain_line(input_folder):
    f = "terrain_line"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POLYLINEZ)
    writer.field("name", "C", size=32)
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_study_area(input_folder):
    f = "study_area"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POLYGONZ)
    writer.field("name", "C", size=32)
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_study_receivers(input_folder):
    f = "study_receiver"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POINTZ)
    writer.field("rec_id", "C", size=10)
    writer.field("x", "N")
    writer.field("y", "N")
    writer.field("z", "N")
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_field_measurements(input_folder):
    f = "field_measurement"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POINTZ)
    writer.field("rec_id", "C", size=10)
    writer.field("x", "N")
    writer.field("y", "N")
    writer.field("z", "N")
    writer.field("dba", "N")
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_ground_area(input_folder):
    f = "ground_zone"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POLYGONZ)
    writer.field("name", "C", size=32)
    writer.field("type", "C", size=32)
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_tree_zone(input_folder):
    f = "tree_zone"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POLYGONZ)
    writer.field("name", "C", size=32)
    writer.field("avg_hgt", "N")
    writer.field("project", "C", size=100)
    return f"{fn}.shp"

def create_building_row(input_folder):
    f = "building_row"
    fn = os.path.join(input_folder, f)
    writer = shapefile2.Writer(fn, shapeType=shapefile2.POLYLINEZ)
    writer.field("name", "C", size=32)
    writer.field("avg_hgt", "N")
    writer.field("bldg_pct", "N")
    return f"{fn}.shp"

if __name__ == '__main__':
    pass


import os
import shapefile

def roadway_separator():
    """Syntax for separating individual roadways in STAMINA format
    
    Returns:
        [String] -- [roadway separator]
    """
    separator = "'L' /\n"
    return separator

def barrier_separator():
    """Syntax for separating individual barriers in STAMINA format
    
    Returns:
        [String] -- [barrier separator]
    """
    separator = "'A' /\n"
    return separator

def _write_roadway_points(line_geom):
    """Converts Polyline to STAMINA syntax 
    
    Arguments:
        line_geom {[arcpy.Polyline]} -- [polyline geometry]
    
    Returns:
        [string] -- [roadway points]
    """
    point_strings = ""
    point_strings += roadway_separator()
    elevation_list = line_geom.z # z values read separately from xy data
    for point_number, point in enumerate(line_geom.points):
        print(point)
        x = round(point[0], 1) 
        y = round(point[1], 1)
        z = round(elevation_list[point_number], 1)            
        point_strings += "'Point{}' {} {} {} 0\n".format(point_number, x, y, z)
        point_strings += roadway_separator()
        return point_strings

def validate_roadway_field(condition):
    if condition == "BUILD":
        return ("auto_bld", "medium_bld", "heavy_bld")
    elif condition == "EXISTING":
        return ("auto_ex", "medium_ex", "heavy_ex")
    elif condition == "NO_BUILD":
        return ("auto_nb", "medium_nb", "heavy_nb")
    else:
        raise ValueError("Feature class must include fields: bld_total, ex_total, nb_total")

def _write_roadways(roadway_feature_class, condition):
    """Writes roadway feature class to STAMINA syntax
    
    Arguments:
        roads_feature_class {String} -- Path to feature class
        condition {String} -- Existing, NoBuild, or Build. Determines fields to use from geospatial template
    Returns:
        [string] -- [roadways]
    """

    roadway_count = len([row for row in shapefile.Reader(roadway_feature_class)])
    with shapefile.Reader(roadway_feature_class) as roadways:
        roadway_string = "2,{}\n".format(roadway_count)
        flds = validate_roadway_field(condition)
        for row in roadways.shapeRecords():
            road = row.record["road_name"]
            speed = row.record["speed"]
            auto = round(row.record[flds[0]], 0)
            medium = round(row.record[flds[1]], 0)
            heavy = round(row.record[flds[2]], 0)
            roadway_string += "{}\n".format(road)
            roadway_string += "CARS {} {}\n".format(auto, speed)
            roadway_string += "MT {} {}\n".format(medium, speed)
            roadway_string += "HT {} {}\n".format(heavy, speed)
            roadway_string += _write_roadway_points(row.shape)
        return roadway_string 

def _write_barrier_points(line_geom, barrier_info=None):
    """Converts Polyline to STAMINA syntax
    
    Arguments:
        line_geom {arcpy.Polyline]} -- [polyline geometry]
    
    Returns:
        [String] -- [barrier points]
    """
    point_strings = ""
    # set constants for barrier defaults TODO: parameterize
    PERTURBATION_INCREMENT = 2 
    NUMBER_OF_PERTURBATIONS = 15    
    BARRIER_INITIAL_HEIGHT = 0
    if barrier_info:
        PERTURBATION_INCREMENT = barrier_info["pert_inc"]
        NUMBER_OF_PERTURBATIONS = barrier_info["pert_num"]    
        BARRIER_INITIAL_HEIGHT = barrier_info["init_hgt"] #TODO: cannot be null, need to account for it
    elevation_list = line_geom.z # z values read separately from xy data
    for point_number, point in enumerate(line_geom.points):
        x = round(point[0], 1) 
        y = round(point[1], 1)
        z = round(elevation_list[point_number], 1) 
        barrier_height = z + BARRIER_INITIAL_HEIGHT
        if point_number == 0:
            point_strings += "'Point{}' {} {} {} {} {} {}\n".format(
                point_number, x, y, barrier_height, z, PERTURBATION_INCREMENT, NUMBER_OF_PERTURBATIONS)                
        else:
            point_strings += "'Point{}' {} {} {} {}\n".format(point_number, x, y, barrier_height, z)
    point_strings += barrier_separator()
    return point_strings

def _write_barriers(barrier_feature_class):
    """Creates string of barrier attributes
    
    Arguments:
        barrier_feature_class {String} -- Path to feature class        
    """
    barrier_count = len([row for row in shapefile.Reader(barrier_feature_class)])
    with shapefile.Reader(barrier_feature_class) as barriers:
        barrier_string = "3,{}\n".format(barrier_count)
        for row in barriers.shapeRecords():
            barrier_info = {
                "pert_inc": row.record["pert_inc"],
                "pert_num": row.record["pert_num"],
                "init_hgt": row.record["init_hgt"]
                }
            name = row.record["name"]
            geometry = row.shape
            barrier_string += "{}\n".format(name)
            barrier_string += _write_barrier_points(geometry, barrier_info)
        return barrier_string

def _write_receivers(receiver_feature_class):
    """Writes receiver feature class to STAMINA format. This method assumes bldg_hgt field is populated.
    For elevations, STAMINA import process will subtract 5 feet (or metric equivalent) from provided Z-value 
    
    Arguments:
        receiver_feature_class {[String]} -- Path to receiver feature class
    """
    receiver_count = len([row for row in shapefile.Reader(receiver_feature_class)])
    receiver_string = "5,{}\n".format(receiver_count)
    receiver_string += "RECEIVERS\n"   
    with shapefile.Reader(receiver_feature_class) as receivers:
        for row in receivers.shapeRecords():         
            rec_id = row.record["rec_id"] # TODO: if no rec id, .dat file does not import properly into TNM 
            bldg_hgt = row.record["bldg_hgt"]            
            x = round(row.shape.points[0][0], 1)
            y = round(row.shape.points[0][1], 1)
            z = round(row.shape.z[0], 1) + bldg_hgt
            receiver_string += "'{}' {} {} {}\n".format(rec_id, x, y, z)
        return receiver_string

def write_stamina_file(file_path, condition, roadways=None, barriers=None, receivers=None):
    """Creates STAMINA file from set of feature classes
    
    Arguments:
        file_path {String} -- Path for output file
        condition {String} -- EXISTING, NOBUILD, or BUILD (case-insensitive)
    
    Keyword Arguments:
        roadways {String} -- Path to roadways feature class (default: {None})
        barriers {String} -- Path to barriers feature class (default: {None})
        receivers {String} -- Path to receivers feature class (default: {None})
    
    Returns:
        [String] -- Path to output STAMINA file
    """
    stamina_string = "1,3\n"
    if roadways:
        stamina_string += _write_roadways(roadways, condition)
    if barriers:
        stamina_string += _write_barriers(barriers)
    if receivers:
        stamina_string += _write_receivers(receivers)
    stamina_string += "7/\n"
    file_obj = open(os.path.join(file_path, "{}.dat".format(condition)), "w")
    file_obj.write(stamina_string)
    return file_path

if __name__ == '__main__':
    # print(_write_roadways(r"C:\Users\Brandon\projects\pytnm\tests\test_files\DATA\existing_roadway.shp", "EXISTING"))
    # print(_write_barriers(r"C:\Users\Brandon\projects\pytnm\tests\test_files\DATA\barrier.shp"))
    # print(_write_receivers(r"C:\Users\Brandon\projects\pytnm\tests\test_files\DATA\receiver.shp"))
    write_stamina_file(r"C:\TNM25",
    "EXISTING", 
    roadways=r"C:\Users\Brandon\projects\pytnm\tests\test_files\DATA\existing_roadway.shp",
    receivers=r"C:\Users\Brandon\projects\pytnm\tests\test_files\DATA\receiver.shp",
    barriers=r"C:\Users\Brandon\projects\pytnm\tests\test_files\DATA\barrier.shp")


import os
from pyproj import CRS, Transformer
# from arcpy import Array
# from arcpy import Describe
# from arcpy import GetCellValue_management
# from arcpy import Polyline
# from arcpy import Polygon
# from arcpy import Point
# from arcpy.da import SearchCursor, UpdateCursor


def project_point(x, y, from_spatial_reference, to_spatial_reference):
    """  """
    new_point = Point(x, y)
    new_point_geom = PointGeometry(new_point, from_spatial_reference)
    projected_point = new_point_geom.projectAs(to_spatial_reference).firstPoint
    new_x = projected_point.X
    new_y = projected_point.Y
    return (new_x, new_y)

# def _get_z_value(coords, raster):
#     coords_string = "{} {}".format(*coords)
#     eval_cell_value = GetCellValue_management(raster, coords_string)
#     try:
#         cell_value = float(eval_cell_value.getOutput(0))
#     except ValueError:
#         return 0
#     return round(cell_value, 1)

# def _update_poly_z(geom, raster, z_factor=1):
#     geom_sr = geom.spatialReference  
#     desc_raster = Describe(raster)    
#     raster_sr = desc_raster.spatialReference 
#     array = Array()
#     for part in geom: # TODO: account for multipart geometry
#         for pnt in part:
#             projected_coords = project_point(pnt.X, pnt.Y, geom_sr, raster_sr)
#             new_z = round(_get_z_value(projected_coords, raster) * z_factor, 1)
#             # new_z = round(pnt.Z * z_factor, 1)
#             pnt.Z = new_z
#             array.add(pnt)
#     return array

# def _round_poly_vertices(geom):
#     array = Array()
#     for part in geom: # TODO: account for multipart geometry
#         for pnt in part:
#             pnt.X = round(pnt.X, 2)
#             pnt.Y = round(pnt.Y, 2)
#             pnt.Z = round(pnt.Z, 2)
#             array.add(pnt)                    
#     return array

# def _convert_z_vertices(geom, z_factor):
#     """Converts z-coordinates to different unit

#     Args:
#         geom ([type]): [description]
#         z_factor ([type]): [description]

#     Returns:
#         [type]: [description]
#     """
#     array = Array()
#     for part in geom: # TODO: account for multipart geometry
#         for pnt in part:
#             pnt.X = pnt.X
#             pnt.Y = pnt.Y
#             pnt.Z = pnt.Z * z_factor
#             array.add(pnt)                    
#     return array

# def _update_z_poly_vertices(geom, xys_to_update):
#     Updates z vales for given coordinate lists, for making z values congruent for identical xy locations
#     array = Array()
#     for part in geom: # TODO: account for multipart geometry
#         for pnt in part:
#             for xyz in xys_to_update:
#                 if pnt.X == xyz[0] and pnt.Y == xyz[1]: 
#                     pnt.X = xyz[0]
#                     pnt.Y = xyz[1]
#                     pnt.Z = xyz[2]
#                 else:
#                     pass
#             array.add(pnt)                    
#     return array

# def _update_point_z(geom, raster, z_factor=1):
#     geom_sr = geom.spatialReference  
#     desc_raster = Describe(raster)    
#     raster_sr = desc_raster.spatialReference
#     pnt = geom.firstPoint # TODO: account for multipart geometry
#     projected_coords = project_point(pnt.X, pnt.Y, geom_sr, raster_sr)
#     new_z = round(_get_z_value(projected_coords, raster) * z_factor, 1)
#     pnt.Z = new_z                  
#     return pnt

# def update_feature_z(fc, raster, z_factor=1, sql=None):
#     desc_fc = Describe(fc)
#     fc_sr = desc_fc.spatialReference
#     with UpdateCursor(fc, ["SHAPE@"], sql) as cursor:
#         for row in cursor:
#             geom = row[0]
#             if geom.type == "polyline":
#                 array = _update_poly_z(geom, raster, z_factor)           
#                 new_geom = Polyline(array, fc_sr, True)
#             elif geom.type == "polygon":
#                 array = _update_poly_z(geom, raster, z_factor)           
#                 new_geom = Polygon(array, fc_sr, True)
#             elif geom.type == "point":
#                 new_geom = _update_point_z(geom, raster, z_factor)
#             else:
#                 try:
#                     raise ValueError('Function requires polyline, polygon, or point feature class')
#                 except ValueError as exc:
#                     print(exc)
#             cursor.updateRow([new_geom])

# def round_coordinates(shp):
#     with UpdateCursor(shp, 'SHAPE@') as ucursor:        
#         for row in ucursor:
#             geom = row[0]
#             array = _round_poly_vertices(geom)
#             new_geom = Polyline(array, None, True)
#             ucursor.updateRow([new_geom])

# def convert_z_coords(shp, z_factor=3.2808399):
#     with UpdateCursor(shp, 'SHAPE@') as ucursor:        
#         for row in ucursor:
#             geom = row[0]
#             array = _convert_z_vertices(geom, z_factor)
#             new_geom = Polyline(array, None, True)
#             ucursor.updateRow([new_geom])

def update_max_z(shp, xys_to_update):
    """Updates z vales for given coordinate lists, for making z values congurent for identical xy locations
    """
    transformer = Transformer.from_crs(from_spatial_reference, to_spatial_reference)
    return transformer.transform(x, y)


# def _get_z_value(coords, raster):
#     coords_string = "{} {}".format(*coords)
#     eval_cell_value = GetCellValue_management(raster, coords_string)
#     try:
#         cell_value = float(eval_cell_value.getOutput(0))
#     except ValueError:
#         return 0
#     return round(cell_value, 1)

# def _update_poly_z(geom, raster, z_factor=1):
#     geom_sr = geom.spatialReference  
#     desc_raster = Describe(raster)    
#     raster_sr = desc_raster.spatialReference 
#     array = Array()
#     for part in geom: # TODO: account for multipart geometry
#         for pnt in part:
#             projected_coords = project_point(pnt.X, pnt.Y, geom_sr, raster_sr)
#             new_z = round(_get_z_value(projected_coords, raster) * z_factor, 1)
#             # new_z = round(pnt.Z * z_factor, 1)
#             pnt.Z = new_z
#             array.add(pnt)
#     return array

# def _round_poly_vertices(geom):
#     array = Array()
#     for part in geom: # TODO: account for multipart geometry
#         for pnt in part:
#             pnt.X = round(pnt.X, 2)
#             pnt.Y = round(pnt.Y, 2)
#             pnt.Z = round(pnt.Z, 2)
#             array.add(pnt)                    
#     return array

# def _update_z_poly_vertices(geom, xys_to_update):
#     """Updates z vales for given coordinate lists, for making z values congruent for identical xy locations
#     """
#     array = Array()
#     for part in geom: # TODO: account for multipart geometry
#         for pnt in part:
#             for xyz in xys_to_update:
#                 if pnt.X == xyz[0] and pnt.Y == xyz[1]: 
#                     pnt.X = xyz[0]
#                     pnt.Y = xyz[1]
#                     pnt.Z = xyz[2]
#                 else:
#                     pass
#             array.add(pnt)                    
#     return array

# def _update_point_z(geom, raster, z_factor=1):
#     geom_sr = geom.spatialReference  
#     desc_raster = Describe(raster)    
#     raster_sr = desc_raster.spatialReference
#     pnt = geom.firstPoint # TODO: account for multipart geometry
#     projected_coords = project_point(pnt.X, pnt.Y, geom_sr, raster_sr)
#     new_z = round(_get_z_value(projected_coords, raster) * z_factor, 1)
#     pnt.Z = new_z                  
#     return pnt

# def update_feature_z(fc, raster, z_factor=1, sql=None):
#     desc_fc = Describe(fc)
#     fc_sr = desc_fc.spatialReference
#     with UpdateCursor(fc, ["SHAPE@"], sql) as cursor:
#         for row in cursor:
#             geom = row[0]
#             if geom.type == "polyline":
#                 array = _update_poly_z(geom, raster, z_factor)           
#                 new_geom = Polyline(array, fc_sr, True)
#             elif geom.type == "polygon":
#                 array = _update_poly_z(geom, raster, z_factor)           
#                 new_geom = Polygon(array, fc_sr, True)
#             elif geom.type == "point":
#                 new_geom = _update_point_z(geom, raster, z_factor)
#             else:
#                 try:
#                     raise ValueError('Function requires polyline, polygon, or point feature class')
#                 except ValueError as exc:
#                     print(exc)
#             cursor.updateRow([new_geom])

# def round_coordinates(shp):
#     with UpdateCursor(shp, 'SHAPE@') as ucursor:        
#         for row in ucursor:
#             geom = row[0]
#             array = _round_poly_vertices(geom)
#             new_geom = Polyline(array, None, True)
#             ucursor.updateRow([new_geom])

# def update_max_z(shp, xys_to_update):
#     """Updates z vales for given coordinate lists, for making z values congurent for identical xy locations

#     Args:
#         shp ([type]): [description]
#     """
#     with UpdateCursor(shp, 'SHAPE@') as ucursor:        
#         for row in ucursor:
#             geom = row[0]
#             array = _update_z_poly_vertices(geom, xys_to_update)
#             new_geom = Polyline(array, None, True)
#             ucursor.updateRow([new_geom])

# def get_all_vertices(shp):
#     vertices = []
#     with SearchCursor(shp, 'SHAPE@') as scursor:
#         for row in scursor:
#             geom = row[0]
#             for part in geom:
#                 for pnt in part:
#                     xyz = [pnt.X, pnt.Y, pnt.Z]
#                     vertices.append(xyz)
#     return vertices

# def identical_pnts(coords):
#     identical_pnts = []
#     for point in coords:
#         identical = 0
#         x = round(point[0], 2)
#         y = round(point[1], 2)
#         for pnt in coords:
#             if round(pnt[0], 2) == x and round(pnt[1], 2) == y:
#                 identical += 1 # all points should be identical to themselves, so one will be minimum
#         if identical > 1:
#             identical_pnts.append(point)
#     return sorted(identical_pnts)

# def identical_set(coords): #produces duplicates but whatever
#     """Returns set of max z values per given identical xy

#     Args:
#         coords (List): List of identical xy coordinates
#     """
#     max_coords = []
#     for coord in coords:
#         identicals = [xy for xy in coords if xy[0] == coord[0] and xy[1] == coord[1]]
#         max_z = max([xyz[2] for xyz in identicals])
#         for identical in identicals:
#             if identical[2] == max_z:
#                 max_coords.append(identical)
#                 break
#             else:
#                 pass
#     return max_coords 


# if __name__ == '__main__':
#     x = 123
#     y = 456
#     print(project_point(x, y, 3857, 4326))


# import arcpy


# def get_count_of_unique_vals(fc, fields):
#     countList = []
#     with arcpy.da.SearchCursor(fc,fields) as cursor:        
#         for row in sorted(cursor):            
#             if not row[0] in countList:
#                 countList.append(row[0])
#     return countList

# def validatespatialreference(fc):
#     desc = arcpy.Describe(fc)
#     mxd = arcpy.mapping.MapDocument("CURRENT")
#     df = arcpy.mapping.ListDataFrames(mxd, "Layers")[0]
#     if desc.spatialReference.name == "Unknown":
#         arcpy.AddWarning("Projection of input {} not defined. \
# 		Setting projection of input polyline to match current data frame...".format(desc.name))
#         arcpy.DefineProjection_management(fc, df.spatialReference)
#     else:
#         arcpy.AddMessage("Projection of input {} is {}. Proceeding...".format(desc.name,desc.spatialReference.name))

# def validate_roadway_fields(roadfc, condition):
#     flds = [f.name for f in arcpy.ListFields(roadfc)]
#     if not ("bld_total" in flds or "ex_total" in flds or "nb_total" in flds):
#         raise ValueError("Feature class must include fields: bld_total, ex_total, nb_total")
#     if condition == "BUILD":
#         return ("road_name", "speed", "auto_bld", "medium_bld", "heavy_bld", "SHAPE@")
#     elif condition == "EXISTING":
#         return ("road_name", "speed", "auto_ex", "medium_ex", "heavy_ex", "SHAPE@")
#     elif condition == "NO_BUILD":
#         return ("road_name", "speed", "auto_nb", "medium_nb", "heavy_nb", "SHAPE@")  


# -*- coding: utf-8 -*-
"""
Purpose: Quickly assess the feasibility and reasonableness 
of a given barrier design, given a reduction design goal,
a reasonable cost, and any other state-specific criteria
"""

import os
import sys
import openpyxl as p


def _xlsx_to_traffic(ws):
    """Converts .xlsx spreadsheet of TNM Traffic to dictionary
	
	Arguments:
		ws {openpyxl.worksheet} -- [description]
	
	Returns:
		[type] -- [description]
	"""
    filter_rows = 7 ##omit these rows at start of spreadsheet    
    traf_ws = ws
    traf_dict = dict([(r[1].value.strip(), ((r[5].value, r[6].value), (r[7].value, r[8].value), (r[9].value, r[10].value)))
                    for r in traf_ws.rows if r[1].value][filter_rows:])
    return traf_dict

def _barriers_to_list(ws):
    """Returns filtered list of barriers
	
	Arguments:
		ws {openpyxl.worksheet} -- [description]
	
	Returns:
		[type] -- [description]
	"""
    barriers_ws = ws
    bars = []
    for row in barriers_ws.rows:
        vals = [cell.value for cell in row]
        bars.append(vals)
    bars_no_header = bars[15:] # clip TNM2.5 header
    barriers = []
    for bar in bars_no_header:
        bar_name, bar_max_height, square_foot_cost = bar[1], bar[4], bar[5]
        x, y, z = bar[13], bar[14], bar[15]
        # check for empty end rows
        if None in (x, y, z):
            break
        if bar_name:
            barriers.append((bar_name, bar_max_height, square_foot_cost, [(x, y, z)]))
            found_bar = True
        else:
            found_bar = False
            if not found_bar:
                barriers[-1][3].append((x, y, z))
    return barriers

def _receivers_to_list(ws):
	receivers_raw = []	
	for row in ws.rows:
		vals = [cell.value for cell in row]
		receivers_raw.append(vals)
	receivers_no_header = receivers_raw[16:] # clip TNM2.5 header
	receivers = []
	for receiver in receivers_no_header:
		rec_id = receiver[1]
		receptors = receiver[3]
		x, y, z = receiver[4], receiver[5], receiver[6]
		height = receiver[7]
		nac_level = receiver[9]
		if None in (x, y, z):
			break
		else:
			receivers.append((rec_id.strip(), receptors, height, nac_level, (x, y, z)))
	return receivers
			
def _rds_to_list(ws):
    """
    Returns filtered list of roads    
    :param ws: 
    :return: road_list
    """    
    roads_ws = ws
    rds = []
    for row in roads_ws.rows:
        vals = [cell.value for cell in row]
        rds.append(vals)
    rds_no_header = rds[15:] # clip TNM2.5 header
    roads = []
    for rd in rds_no_header:
        rd_name = rd[1]
        rd_width = rd[2]
        pnt = rd[4].strip()
        x, y, z = rd[6], rd[7], rd[8]
        # check for empty end rows
        if None in (x, y, z):
            break
        if rd_name:
            roads.append([rd_name.strip(), rd_width, [[pnt,[x, y, z]]]])
            found_road = True
        else:
            found_road = False
            if not found_road:
                roads[-1][2].append([pnt,[x, y, z]])
    return roads

def _validate_xlsx(xlsx):
	pass

def _read_roadways(xlsx):
	workbook = p.load_workbook(xlsx)
	try:
		roadways_worksheet = [ws for ws in workbook.worksheets if ws['B5'].value == 'INPUT: ROADWAYS'][0]
	except IndexError:
		print('Roadways worksheet not found')
	return(_rds_to_list(roadways_worksheet))

def _read_traffic(xlsx):
	workbook = p.load_workbook(xlsx)
	try:
		traffic_worksheet = [ws for ws in workbook.worksheets if ws['B5'].value == 'INPUT: TRAFFIC FOR LAeq1h Volumes'][0]
	except IndexError:
		print('Traffic worksheet not found')
	return(_xlsx_to_traffic(traffic_worksheet))

def _read_receivers(xlsx):
	workbook = p.load_workbook(xlsx)
	try:
		receiver_worksheet = [ws for ws in workbook.worksheets if ws['B5'].value == 'INPUT: RECEIVERS'][0]
	except IndexError:
		print('Receiver worksheet not found')
	return(_receivers_to_list(receiver_worksheet))

def _read_barriers(xlsx):
	workbook = p.load_workbook(xlsx)
	try:
		barrier_worksheet = [ws for ws in workbook.worksheets if ws['B5'].value == 'INPUT: BARRIERS'][0]
	except IndexError:
		print('Barrier worksheet not found')
	return(_barriers_to_list(barrier_worksheet))

class VehicleClassification(object):
	def __init__(self, volume, speed):
		self.volume = volume
		self.speed = speed

class Auto(VehicleClassification):
	def __repr__(self):
		return 'Auto'

class Medium(VehicleClassification):
	def __repr__(self):
		return 'Medium'

class Heavy(VehicleClassification):
	def __repr__(self):
		return 'Heavy'

class Traffic(object):
	"""Represents Traffic TNM object
	
	Arguments:
		object {[type]} -- [description]
	
	Returns:
		[type] -- [description]
	"""
	def __init__(self, auto, medium, heavy, bus=None, motorcycle=None):
		self.auto = auto
		self.medium = medium
		self.heavy = heavy
		self.bus = bus
		self.motorcycle = motorcycle
	def __repr__(self):
		return self.auto

class Roadway(object):
	"""Represents Roadway TNM object
	
	Arguments:
		object {[type]} -- [description]
	
	Returns:
		[type] -- [description]
	"""
	def __init__(self, name, width, points, traffic=None):
		self.name = name
		self.width = width
		self.points = points
		self.traffic = traffic
		self.geometry = None
	def __repr__(self):
		return self.name


class Receiver(object):
	"""Represents Receiver TNM object
	
	Arguments:
		object {[type]} -- [description]
	"""
	def __init__(self, rec_id, receptors, height, nac_level, geometry, nac=None):
		self.rec_id = rec_id
		self.receptors = receptors
		self.height = height
		self.nac_level = nac_level
		self.geometry = geometry
	def __repr__(self):
		return self.rec_id

class Barrier(object):
	"""Represents Barrier TNM object
	
	Arguments:
		object {[type]} -- [description]
	
	Returns:
		[type] -- [description]
	"""
	def __init__(self, name, max_height, geometry, square_foot_cost=0):
		self.name = name
		self.max_height = max_height
		self.geometry = geometry
		self.square_foot_cost = square_foot_cost
	def __repr__(self):
		return self.name

class TerrainLine(object):
	pass

class BuildingRow(object):
	pass

class Model(object):
	"""[summary]
	
	>> build_run = pytnm.Model(string(.xlsx))
	>> print(build_run.info)
	>> '(('Build Run', 'Project', 'Company', 'Analyst', 'Date'))
	>> roadways = build_run.roadways
	>> roadway = roadways[0]
	>> print(roadway.name)
	>> 'I-10 On Ramp 1'
	>> print(roadway.width)
	>> 22
	>> print(roadway.geometry)
	>> ((123456, 4355422, 12), (123456, 4355422, 15), (123456, 4355422, 12))
	>> print(roadway.traffic)
	>> (125, 10, 5, 3, 1)
	>> print(roadway.traffic.auto)
	>> 125
	>> print(roadway.traffic.auto.speed)
	>> 45

	Arguments:
		object {[type]} -- [description]
	"""

	def __init__(self, xlsx):
		self.TRAFFIC = _read_traffic(xlsx)
		self._roads = _read_roadways(xlsx)
		self._receivers = _read_receivers(xlsx)
		self._barriers = _read_barriers(xlsx)
	
	@property
	def barriers(self):
		barriers = []
		for barrier in self._barriers:
			name = barrier[0]
			max_height = barrier[1]
			geometry = barrier[2]
			barriers.append(Barrier(name, max_height, geometry))		
		return barriers

	@property
	def receivers(self):
		receivers = []		
		for receiver in self._receivers:
			rec_id = receiver[0]
			receptors = receiver[1]			
			height = receiver[2]
			nac_level = receiver[3]
			x, y, z = receiver[4]
			receivers.append(Receiver(rec_id, receptors, height, nac_level, geometry=(x, y, z)))
		return receivers

	@property	
	def roadways(self):
		roads = []
		for road in self._roads:
			name = road[0]
			width = road[1]
			points = road[2]
			autos = self.TRAFFIC[name][0]
			medium = self.TRAFFIC[name][1]
			heavy = self.TRAFFIC[name][2]
			traffic = Traffic(Auto(autos[0], autos[1]), Medium(medium[0], medium[1]), Heavy(heavy[0], heavy[1]))
			roads.append(Roadway(name, width, points, traffic))
		return roads		

class Analysis(object):
	"""Barrier analysis class
	
	This class will contain several read-only properties
	that will aid in documenting the feasibility and 
	reasonableness of a given barrier design modeled in the FHWA
	TNM 2.5. Information is pulled from an ".xlsx" wb of any name 
	with the assumption that the Barrier Design Table results and
	Sound Level Results tables are copied AS-IS directly from TNM 2.5
	to two separate worksheets in the ".xlsx" workbook.
	
	:param wbname: ".xlsx" file on disk (i.g. "Barriers.xlsx")
	:param barsheet: worksheet in ".xlsx" containing Barrier Design Table
	:param sndsheet: "worksheet in ".xlsx" containing Sound Level Results
	:param barriercost: cost of barrier design, as reported by TNM 2.5
	"""

	def __init__(self, wbname, barsheet, sndsheet, barcost=0):
		self._wbname = wbname
		self._barsheet = barsheet
		self._sndsheet = sndsheet
		if barsheet == sndsheet:
			raise ValueError("Worksheets cannot be identical!")
		self.barriercost = barcost
		self._barrecs = self._wbhandler(barsheet)
		self._sndrecs = self._wbhandler(sndsheet)
		"""Receivers listed in the Sound Results worksheet"""
		self.snd_rec_list = [
                                  (i[0].value, i[2].value, i[8].value) 
                                      for i in self._sndrecs
                               ]
		self.impacted_recs = [(i[0], i[1]) for i in self.recs_analysis 
                                      if i[4] != " ----"
                                ]
		self.impact_num = sum([tup[1] for tup in self.impacted_recs])
		self.du_analysis = sum([tup[1] for tup in self.recs_analysis])
		self.benefitted =  [
                                  (i[0], i[1]) for i in self.recs_analysis 
                                      if i[2] >= 5
                              ]
		self.benefit_num = sum([tup[1] for tup in self.benefitted])
		self.ben_and_imp = [
                                  (i[0], i[1]) for i in self.recs_analysis 
                                      if i[2] >= 5 and i[4] != " ----"
                               ]
		self.ben_and_imp_num = sum([tup[1] for tup in self.ben_and_imp])
		self.reas_red_recs = [
                                  (i[0], i[1]) for i in self.recs_analysis
                                      if i[2] >= i[3]
                                ]
		self.reas_red_num = sum([tup[1] for tup in self.reas_red_recs])
	def _wbhandler(self, sht):
		"""
		Load Excel workbook and generate appropriate receiver list
		depending on the type of results contained in the worksheet
		"""
		wb = p.load_workbook(self._wbname)
		ws = wb.get_sheet_by_name(sht)
		if not ws:
			raise ValueError("Excel worksheet not found!" + 
                                       "Is it spelled correctly?")
		if sht == self._barsheet:
			lastrow = ws.max_row
			#assumes copied from TNM bar sound results in cell A1
			datarange = 'B18:L{}'.format(lastrow)
		elif sht == self._sndsheet:
			#omit last 8 rows in snd results
			lastrow = ws.max_row - 8
			datarange = 'B20:N{}'.format(lastrow)
		reclist = list(ws.iter_rows(range_string=datarange))
		return reclist
	@property		
	def recs_analysis(self):
		"""
		List receivers, DU, barrier reduction value, reduction goal, and
		impact status in chosen barrier analysis
		"""
		barreclist = self._barrecs
		sndreclist = self._sndrecs
		r = []
		for barrec in barreclist:
			rec = barrec[0].value
			barred = barrec[3].value
			redgoal = barrec[4].value
			for sndrec in sndreclist:
				if rec == sndrec[0].value:
					du = sndrec[2].value
					impstat = sndrec[8].value
					r.append((rec, du, barred, redgoal, impstat))
				else:
					pass
		return sorted(r)
	@property
	def benefits(self):
		"""
		Return tuple of benefited receiver/receptor counts
		"""
		return [(r[0], r[1]) for r in self.recs_analysis
					if r[2] >= 5]
	@property
	def perc_imp_benefitted(self):
		"""
		Return percentage of total receivers in barrier analysis that
		receive a 5 dBA reduction
		"""
		perc = float(self.ben_and_imp_num) / float(self.impact_num)
		return "{:.0%}".format(perc)
	@property
	def perc_ben_reasonable(self):
		"""
		Return percentage of benefits in barrier analysis that
		receive an reasonable reduction
		"""
		if self.benefit_num == 0:
			perc = 0
		else:
			perc = float(self.reas_red_num) / float(self.benefit_num)
		return "{:.0%}".format(perc)
	@property
	def feasible(self):
		"""
		Barrier design is feasible if 75% or more of impacted receivers
		receive a 5dBA or more noise reduction
		"""
		perc_crit = self.impact_num * 0.75
		if self.benefit_num >= perc_crit:
			return True
		else:
			return False
	@property
	def reasonable(self):
		"""
		Barrier design is reasonable if 80% or more of benefitted receivers
		receive a 8dBA or more noise reduction
		"""
		perc_crit = self.benefit_num * 0.80
		if perc_crit == 0:
			return False
		elif self.reas_red_num >= perc_crit:
			return True
		else:
			return False
	def cost_per_benefit(self):
		"""
		If a barrier cost has been specified, returns the cost per 
		benefitted receptor based on the policy-specific allowable cost
		"""
		if self.barriercost <= 0:
		    return 0
		else:
		    return self.barriercost / self.benefit_num
	def report(self):
		"""
		Report of results. Useful for report writing or debugging.
		"""
		pass
		# print("Receivers/receptors in barrier analysis: {} ({})".format(\
        #                       len(self.recs_analysis), self.du_analysis))
		# print "Impacts in barrier analysis: {} ({})".format(\
        #                       len(self.impacted_recs), self.impact_num)
		# print "Benefits in barrier analysis: {} ({})".format(\
        #                       len(self.benefitted), self.benefit_num)
		# print "Number of impacts receiving 5dBA reduction: {}".format(\
        #                       self.ben_and_imp_num)
		# print "Impacts (%) receiving 5dBA reduction: {}".format(\
        #                       self.perc_imp_benefitted)
		# print "Benefits receiving 8dBA reduction: {} ({})".format(\
        #                       len(self.reas_red_recs), self.reas_red_num) 
		# print "Benefits (%) receiving reasonable reduction: {}".format(\
        #                       self.perc_ben_reasonable) 
		# print "Barrier design is feasible: {}".format(\
        #                       self.feasible)
		# print "Barrier design is reasonable: {}".format(\
        #                       self.reasonable)

# class GAAnalysis(Analysis):
#     def __init__(self, wbname, barsheet, sndsheet, barcost=0):
#         super(GAAnalysis, self).__init__(wbname, barsheet, sndsheet, barcost=0)
# 	@property
# 	def feasible(self):
# 		"""
#         Barrier design is feasible if one or more of impacted receivers
#         receive a 5dBA or more noise reduction
#         """
#         if self.ben_and_imp_num >= 1:
#             return True
#         return False

# 	@property
# 	def reasonable(self):
# 		"""
#         Barrier design is reasonable if 80% or more of benefitted receivers
#         receive a 8dBA or more noise reduction
#         """
# 		perc_crit = self.benefit_num * 0.80
# 		if perc_crit == 0:
# 			return False
# 		elif self.reas_red_num >= perc_crit:
# 			return True
# 		else:
# 			return False

if __name__ == '__main__':
	os.chdir(os.path.dirname(__file__))
	model_xlsx = '../files/tnm_model.xlsx'
	build_model = Model(model_xlsx)
	roadways = build_model.roadways	
	# for roadway in roadways:
	# 	print(roadway.traffic.auto.speed)
	# 	print(roadway.traffic.medium.speed)
	# 	print(roadway.traffic.heavy.speed)
	# receivers = build_model.receivers	
	# for receiver in receivers:
	# 	print(receiver)
	barriers = build_model.barriers
	print(barriers)
	for barrier in barriers:
		print(barrier)



import pathlib
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="pytnm",
    version="1.0.4",
    description="FHWA TNM 2.5 Geospatial Toolkit",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/fstraw/pytnm",
    author="Brandon Batt",
    author_email="brbatt@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=[
		"pytnm",
		"pytnm.utils",
		"pytnm.utils.report"
	],
    install_requires=["openpyxl", "pyshp"],
)


import itertools
import openpyxl
import shapefile


class Analysis(object):
    pass

class Report(object):
    """ Reports receiver results based on shapefile of known schema
    """       

    def __init__(self, receivers, common_noise_environment=None):
        self.shp = receivers
        self.common_noise_env = common_noise_environment
        self.SENSITIVE_QUERY = ('A', 'B', 'C', 'D', 'E')   

    def _get_field_index(self, fields):
        flds = [fld for fld in fields if fld[0] != 'DeletionFlag'] # removed hidden field for proper indexing
        field_map = {}
        for index, field in enumerate(flds):
            field_name = field[0]
            field_map[field_name] = index    
        return field_map

    def _field_map(self):
        return self._get_field_index(shapefile.Reader(self.shp).fields)

    def all_receivers(self):
        with shapefile.Reader(self.shp) as sreader:
            if self.common_noise_env:
                cne = self._field_map()['assoc_bar'] 
                return [record for record in sreader.records() if record[cne] == self.common_noise_env]
            else:    
                return [record for record in sreader.records()]

    def nac_b_receivers(self):
        nac_cat = self._field_map()['nac_cat']                       
        return [record for record in self.all_receivers() if record[nac_cat] == 'B']

    def nac_c_receivers(self):
        nac_cat = self._field_map()['nac_cat']                       
        return [record for record in self.all_receivers() if record[nac_cat] == 'C']

    def nac_d_receivers(self):
        nac_cat = self._field_map()['nac_cat']                       
        return [record for record in self.all_receivers() if record[nac_cat] == 'D']

    def nac_e_receivers(self):
        nac_cat = self._field_map()['nac_cat']                       
        return [record for record in self.all_receivers() if record[nac_cat] == 'E']

    def nac_b_receptors_total(self):
        du = self._field_map()['du']
        nac_cat = self._field_map()['nac_cat']                       
        return sum([row[du] for row in self.all_receivers() if row[nac_cat] == 'B'])

    def nac_c_receptors_total(self):
        du = self._field_map()['du']
        nac_cat = self._field_map()['nac_cat']                       
        return sum([row[du] for row in self.all_receivers() if row[nac_cat] == 'C'])

    def nac_d_receptors_total(self):
        du = self._field_map()['du']
        nac_cat = self._field_map()['nac_cat']                       
        return sum([row[du] for row in self.all_receivers() if row[nac_cat] == 'D'])

    def nac_e_receptors_total(self):
        du = self._field_map()['du']
        nac_cat = self._field_map()['nac_cat']                       
        return sum([row[du] for row in self.all_receivers() if row[nac_cat] == 'E'])

    def sensitive_receivers(self):
        nac_cat = self._field_map()['nac_cat']                       
        return [record for record in self.all_receivers() if record[nac_cat] in self.SENSITIVE_QUERY]

    def sensitive_receptors_total(self):
        du = self._field_map()['du']
        nac_cat = self._field_map()['nac_cat']        
        return sum([row[du] for row in self.all_receivers() if row[nac_cat] in self.SENSITIVE_QUERY])

    def benefited_receptors(self):
        du = self._field_map()['du']
        bar_reduct = self._field_map()['bar_reduct']
        return sum([record[du] for record in self.sensitive_receivers() if record[bar_reduct] >= 5])

    def receptors_total(self):
        du = self._field_map()['du']
        return sum([row[du] for row in self.all_receivers()])

    def impacted_receivers(self):
        impacts = list(
            itertools.chain(
            self.build_b_nac_impacts(),
            self.build_c_nac_impacts(),
            self.build_d_nac_impacts(),
            self.build_e_nac_impacts(),
            self.build_b_substantial_impacts(),
            self.build_c_substantial_impacts(),
            self.build_d_substantial_impacts(),
            self.build_e_substantial_impacts()
            )
        )            
        return impacts

    def impacted_total(self):
        du = self._field_map()['du']
        return sum([impact[du] for impact in self.impacted_receivers()])

    def build_b_impacted_total(self):
        nac_cat = self._field_map()['nac_cat']
        du = self._field_map()['du']
        return sum([impact[du] for impact in self.impacted_receivers() if impact[nac_cat] == 'B'])

    def build_c_impacted_total(self):
        nac_cat = self._field_map()['nac_cat']
        du = self._field_map()['du']
        return sum([impact[du] for impact in self.impacted_receivers() if impact[nac_cat] == 'C'])

    def build_d_impacted_total(self):
        nac_cat = self._field_map()['nac_cat']
        du = self._field_map()['du']
        return sum([impact[du] for impact in self.impacted_receivers() if impact[nac_cat] == 'D'])

    def build_e_impacted_total(self):
        nac_cat = self._field_map()['nac_cat']
        du = self._field_map()['du']
        return sum([impact[du] for impact in self.impacted_receivers() if impact[nac_cat] == 'E'])

    def build_b_nac_impacts(self):        
        nac_cat = self._field_map()['nac_cat']
        bld_snd = self._field_map()['bld_snd']
        return [row for row in self.sensitive_receivers() if (row[nac_cat] == 'B' and row[bld_snd] >= 66)]

    def build_c_nac_impacts(self):        
        nac_cat = self._field_map()['nac_cat']
        bld_snd = self._field_map()['bld_snd']
        return [row for row in self.sensitive_receivers() if (row[nac_cat] == 'C' and row[bld_snd] >= 66)]

    def build_d_nac_impacts(self):        
        nac_cat = self._field_map()['nac_cat']
        bld_snd = self._field_map()['bld_snd']
        return [row for row in self.sensitive_receivers() if (row[nac_cat] == 'D' and row[bld_snd] >= 51)]

    def build_e_nac_impacts(self):        
        nac_cat = self._field_map()['nac_cat']
        bld_snd = self._field_map()['bld_snd']
        return [row for row in self.sensitive_receivers() if (row[nac_cat] == 'E' and row[bld_snd] >= 71)]

    def build_b_substantial_impacts(self):        
        nac_cat = self._field_map()['nac_cat']
        bld_snd = self._field_map()['bld_snd']
        ex_snd = self._field_map()['ex_snd']
        return [row for row in self.sensitive_receivers() if (row[nac_cat] == 'B' and (row[bld_snd] - row[ex_snd] >= 15))]

    def build_c_substantial_impacts(self):        
        nac_cat = self._field_map()['nac_cat']
        bld_snd = self._field_map()['bld_snd']
        ex_snd = self._field_map()['ex_snd']
        return [row for row in self.sensitive_receivers() if (row[nac_cat] == 'C' and (row[bld_snd] - row[ex_snd] >= 15))]

    def build_d_substantial_impacts(self):        
        nac_cat = self._field_map()['nac_cat']
        bld_snd = self._field_map()['bld_snd']
        ex_snd = self._field_map()['ex_snd']
        return [row for row in self.sensitive_receivers() if (row[nac_cat] == 'D' and (row[bld_snd] - row[ex_snd] >= 15))]

    def build_e_substantial_impacts(self):  
        nac_cat = self._field_map()['nac_cat']
        bld_snd = self._field_map()['bld_snd']
        ex_snd = self._field_map()['ex_snd']
        return [row for row in self.sensitive_receivers() if (row[nac_cat] == 'E' and (row[bld_snd] - row[ex_snd] >= 15))]

    def existing_minimum(self):
        ex_snd = self._field_map()['ex_snd']
        return min([row[ex_snd] for row in self.sensitive_receivers()])

    def existing_maximum(self):
        ex_snd = self._field_map()['ex_snd']
        return max([row[ex_snd] for row in self.sensitive_receivers()])

    def nobld_minimum(self):
        nobld_snd = self._field_map()['nobld_snd']
        return min([row[nobld_snd] for row in self.sensitive_receivers()])

    def nobld_maximum(self):
        nobld_snd = self._field_map()['nobld_snd']
        return max([row[nobld_snd] for row in self.sensitive_receivers()])

    def bld_minimum(self):
        bld_snd = self._field_map()['bld_snd']
        return min([row[bld_snd] for row in self.sensitive_receivers()])

    def bld_maximum(self):
        bld_snd = self._field_map()['bld_snd']
        return max([row[bld_snd] for row in self.sensitive_receivers()])
    
    def bld_average_change(self):
        ex_snd = self._field_map()['ex_snd']
        bld_snd = self._field_map()['bld_snd']
        diff_list = [row[bld_snd] - row[ex_snd] for row in self.sensitive_receivers()]
        return round(sum(diff_list) / len(diff_list), 1)

    def nobld_average_change(self):
        ex_snd = self._field_map()['ex_snd']
        nobld_snd = self._field_map()['nobld_snd']
        diff_list = [row[nobld_snd] - row[ex_snd] for row in self.sensitive_receivers()]
        return round(sum(diff_list) / len(diff_list), 1)

    def summary(self):
        narrative = f"A total of {self.sensitive_receptors_total()} noise sensitive receptors were analyzed. "
        narrative += f"Existing noise levels range from {self.existing_minimum()} to {self.existing_maximum()} dB(A) at {self.sensitive_receptors_total()} receptors. "
        narrative += f"No Build noise levels would range from {self.nobld_minimum()} to {self.nobld_maximum()} dB(A). "
        narrative += f"Build noise levels would range from {self.bld_minimum()} to {self.bld_maximum()} dB(A). "        
        narrative += f"Noise is predicted to change by an average of {self.nobld_average_change()} dB(A) under the No Build Alternative. "
        narrative += f"Noise is predicted to change by an average of {self.bld_average_change()} dB(A) under the Build Alternative. "
        if self.impacted_total() == 0:
            narrative += f"No receptors would be impacted under the Build Alternative."
        else:
            narrative += f"A total of {self.impacted_total()} receptors would be impacted under the Build Alternative; therefore, noise abatement was considered."
        return narrative


def create_florida_barrier_summary(xlsx, bar_length, cost_per_sq_ft):
    """ Generate Florida barrier summary based on Sound Results table for *each* barrier
    analysis in TNM Run. 
    
    Arguments:
        xlsx {String} -- Path to xlsx
    """
    xlsx = openpyxl.load_workbook(xlsx)
    barrier_summaries = xlsx.worksheets
    empty_row = [None] * 14 # empty cells, TNM standard output
    increments = [8, 10, 12, 14, 16, 18, 20, 22] #feet
    barrier_summary = []
    for idx, bar_hgt in enumerate(increments):
        table_rows = [row for row in barrier_summaries[idx].rows][19:]
        results = []
        for row in table_rows:
            cells = [cell.value for cell in row]
            if cells != empty_row:
                results.append(cells)
            else:
                break
        impacted_receptors = [row for row in results if row[9] != ' ----']
        impacted_receptors_count = sum([row[3] for row in impacted_receptors])
        min_benefit = sum([row[3] for row in impacted_receptors if 5 <= row[11] <= 5.9])
        mid_benefit = sum([row[3] for row in impacted_receptors if 6 <= row[11] <= 6.9])
        max_benefit = sum([row[3] for row in impacted_receptors if 7 <= row[11]])
        benefited_receptors = [row for row in results if row[11] >= 5]
        benefited_receptor_count = sum([row[3] for row in benefited_receptors])
        impacted_benefited_receptor_count = sum([row[3] for row in impacted_receptors if row[11] >= 5])
        impacted_not_benefited_receptor_count = benefited_receptor_count - impacted_benefited_receptor_count
        estimated_cost = round(bar_length * bar_hgt * cost_per_sq_ft)
        try:
            average_benefit_reduction = round(sum([row[11] for row in benefited_receptors]) / len(benefited_receptors), 1)
            cost_per_benefit = round(estimated_cost / benefited_receptor_count)
        except ZeroDivisionError:
            average_benefit_reduction = 0        
            cost_per_benefit = 0
        barrier_summary.append((
            bar_hgt, '{:,.0f}'.format(bar_length), impacted_receptors_count, min_benefit, mid_benefit, max_benefit, 
            impacted_benefited_receptor_count, impacted_not_benefited_receptor_count, benefited_receptor_count, 
            average_benefit_reduction, estimated_cost, cost_per_benefit))
    return barrier_summary

def summarize_cne(receivers, cne):
    r = Report(receivers, cne)    
    return (
        cne, 
        'LOCATION', 
        r.nac_b_receptors_total(), 
        r.nac_c_receptors_total(),
        r.nac_e_receptors_total(),
        r.existing_minimum(),
        r.existing_maximum(),
        r.nobld_minimum(),
        r.nobld_maximum(),
        r.bld_minimum(),
        r.bld_maximum(),
        r.build_b_impacted_total(),
        r.build_c_impacted_total(),
        r.build_e_impacted_total(),
        'WARRANTED'
    )

def create_cne_summary_table(receivers, cne_list, xlsx):
    wb = openpyxl.Workbook()
    ws = wb.active
    hdrs = ('CNE', 'Location', 'NAC B', 'NAC C', 'NAC E', 'Min', 'Max', 'Min', 'Max', 'Min', 'Max', 'NAC B', 'NAC C', 'NAC E', 'Abatement Warranted')
    ws.append(hdrs)
    for cne in cne_list:
        cne_summary = summarize_cne(receivers, cne)
        ws.append(cne_summary)    
    wb.save(filename=xlsx)

def create_barrier_summary_table(barrier_summary, xlsx):
    """Generate .xlsx file of barrier results
    
    Arguments:
        barrier_summary {}
        xlsx {String} -- Full output path
    """
    wb = openpyxl.Workbook()
    ws = wb.active
    hdrs = ("Barrier Height", "Barrier Length", "Impacted Receptors", 
    "5 - 5.9 dB(A)", "6 - 6.9 dB(A)", "7+", "Impacted", "Not Impacted", 
    "Total", "Average Benefited Reduction", "Estimated Cost", "Cost per Benefit", "Feasible and Reasonable")
    ws.append(hdrs)
    for barrier_design in barrier_summary:
        ws.append(barrier_design)
    wb.save(filename=xlsx)


"""

Convert excel to list or tuple

"""
import os
import shapefile
import openpyxl


def bar_point_id_xy_dict(wbname, ws='Barriers'):
    wb = openpyxl.load_workbook(wbname)
    barriers_ws = wb[ws]
    bars = []
    for row in barriers_ws.rows:
        vals = [cell.value for cell in row]
        bars.append(vals)
    bars_no_header = bars[15:] #clip TNM2.5 header
    return {row[11]: (row[13], row[14], row[15]) for row in bars_no_header}

def get_last_barrier_seg_pointid(wbname, ws='Barriers'):
    """
    Returns last point for each barrier - temporary hack
    :param wbname:
    :param ws:
    :return:
    """
    wb = openpyxl.load_workbook(wbname)
    barriers_ws = wb[ws]
    bars = []
    lastpointdict = {}
    for row in barriers_ws.rows:
        vals = [cell.value for cell in row]
        bars.append(vals)
    bars_no_header = bars[15:] #clip TNM2.5 header
    for row in bars_no_header:
        barname = row[1]
        lastindicator = row[17]
        if barname:
            currentbar = barname
        if lastindicator == None:
            lastpointdict[currentbar] = row[11]
    return lastpointdict

def rds_to_list(wbname, ws='Roads'):
    """
    Returns filtered list of roads
    :param wb: TNM excel template (.xlsx)
    :param ws: 
    :return: 
    """
    wb = openpyxl.load_workbook(wbname)
    roads_ws = wb[ws]
    rds = []
    for row in roads_ws.rows:
        vals = [cell.value for cell in row]
        rds.append(vals)
    rds_no_header = rds[15:] #clip TNM2.5 header
    roads = []
    for rd in rds_no_header:
        rd_name, rd_width = rd[1], rd[2]
        x, y, z = rd[6], rd[7], rd[8]
        #check for empty end rows
        if None in (x, y, z):
            break
        if rd_name:
            roads.append([rd_name, rd_width, [[x, y, z]]])
            found_road = True
        else:
            found_road = False
            if not found_road:
                roads[-1][2].append([x, y, z])
    return roads

def barriers_to_list(wbname, ws='Barriers'):
    """
    Returns filtered list of barriers
    :param wb: TNM excel template (.xlsx)
    :param ws: 
    :return: 
    """
    wb = openpyxl.load_workbook(wbname)
    barriers_ws = wb[ws]
    bars = []
    for row in barriers_ws.rows:
        vals = [cell.value for cell in row]
        bars.append(vals)
    bars_no_header = bars[15:] #clip TNM2.5 header
    barriers = []
    for bar in bars_no_header:
        bar_name, bar_max_height, square_foot_cost = bar[1], bar[4], bar[5]
        x, y, z = bar[13], bar[14], bar[15]
        #check for empty end rows
        if None in (x, y, z):
            break
        if bar_name:
            barriers.append([bar_name, bar_max_height, square_foot_cost, [[x, y, z]]])
            found_bar = True
        else:
            found_bar = False
            if not found_bar:
                barriers[-1][3].append([x, y, z])
    return barriers

def group_barrier_segments_by_height(wbname, ws='Barrier_Segments'):
    """
    Parses barrier into individual segment lengths, based on Barrier Segments output (TNM)
    :param wbname:
    :param ws:
    :return:
    """
    wb = openpyxl.load_workbook(wbname)
    barriers_ws = wb[ws]
    bars = []
    for row in barriers_ws.rows:
        vals = [cell.value for cell in row]
        bars.append(vals)
    bars_no_header = bars[15:] #clip TNM2.5 header
    barriers = []
    #coupled - consider revisiting approach
    XYZ_DICT = bar_point_id_xy_dict(wbname)
    LAST_POINTS_DICT = get_last_barrier_seg_pointid(wbname) #consider revising approach
    i = 0
    while i < len(bars_no_header): #avoid index error
        bar_name, bar_height = bars_no_header[i][1], bars_no_header[i][7]
        try:
            pointid1 = bars_no_header[i][4]
            x1, y1, z1 = XYZ_DICT[pointid1]  # should return tuple
            pointid2 = bars_no_header[i+1][4]
            x2, y2, z2 = XYZ_DICT[pointid2]
        except IndexError: #account for end of list
            pointid2 = LAST_POINTS_DICT[current_bar] # should be assigned by the time this error throws
            x2, y2, z2 = XYZ_DICT[pointid2]
            break
        except KeyError: #check for empty row
            bar_height = bars_no_header[i-1][7]
            pointid1 = bars_no_header[i-1][4]
            x1, y1, z1 = XYZ_DICT[pointid1]  # should return tuple
            pointid2 = LAST_POINTS_DICT[current_bar]  # should be assigned by the time this error throws
            x2, y2, z2 = XYZ_DICT[pointid2]
        if not pointid1:
            bar_height = bars_no_header[i-1][7]
            pointid1 = bars_no_header[i-1][4]
            pointid2 = LAST_POINTS_DICT[current_bar]
            x1, y1, z1 = XYZ_DICT[pointid1]
            x2, y2, z2 = XYZ_DICT[pointid2]
        if bar_name:
            current_bar = bar_name
        if bar_height == 0:
            i+=1
            continue
        barriers.append([current_bar, pointid1, bar_height, [[x1, y1, z1], [x2, y2, z2]]])
        i+=1
    return barriers

def rds_list_to_shape(rdslist, outputshp, traf_dict=None):
    w = shapefile.Writer(shapeType=shapefile.POLYLINEZ)
    w.field('Rd_Name', 'C', size=32)
    w.field('Width', 'N')
    w.field('Auto', 'N')
    w.field('Medium', 'N')
    w.field('Heavy', 'N')
    w.field('MedPct', 'N')
    w.field('HvyPct', 'N')
    w.field('Speed', 'C')
    for rd in rdslist:
        w.line(parts=[rd[2]], shapeType=13)
        if not traf_dict:
            w.record(rd[0], int(rd[1]), 0, 0, 0, 0, 0, 0)
        else:
            auto = traf_dict[rd[0]][0]
            medium = traf_dict[rd[0]][1]
            heavy = traf_dict[rd[0]][2]
            medpct = traf_dict[rd[0]][3]
            hvypct = traf_dict[rd[0]][4]
            speed = traf_dict[rd[0]][5]
            w.record(rd[0], int(rd[1]), auto, medium, heavy, medpct, hvypct, speed)
    w.save(outputshp)
    return outputshp

def bars_list_to_shape(barslist, outputshp):
    w = shapefile.Writer(shapefile.POLYLINEZ)
    w.field('Bar_Name', 'C', size=32)
    w.field('MaxHeight', 'N')
    w.field('CostPerSF', 'N')
    for bar in barslist:
        w.line(parts=[bar[3]], shapeType=13)
        w.record(bar[0], int(bar[1]), bar[2])
    w.save(outputshp)
    return outputshp

def bar_segs_list_to_shape(barslist, outputshp):
    w = shapefile.Writer(shapefile.POLYLINEZ)
    w.field('Bar_Name', 'C', size=32)
    w.field('PointID', 'C', size=10)
    w.field('AVGHEIGHT', 'N')
    for bar in barslist:
        w.line(parts=[bar[3]], shapeType=13)
        w.record(bar[0], bar[1], int(bar[2]))
    w.save(outputshp)
    return outputshp

def append_tnm_traffic(wbname, ws='Traffic'):
    wb = openpyxl.load_workbook(wbname)
    traffic_ws = wb[ws]
    traf = []
    for row in traffic_ws.rows:
        vals = [cell.value for cell in row]
        traf.append(vals)
    traf_no_header = traf[14:] #clip TNM2.5 header
    traffic_vols = []
    for traffic in traf_no_header:
        rd_name, auto, medium, heavy, speed = traffic[1], traffic[5], traffic[7], traffic[9], traffic[6]
        medpct = 0
        hvypct = 0
        if rd_name:
            if medium:
                medpct = round(float(medium) / (auto + medium + heavy) * 100, 1)
            if heavy:
                hvypct = round(float(heavy) / (auto + medium + heavy) * 100, 1)
            traffic_vols.append([rd_name, auto, medium, heavy, medpct, hvypct, speed])            
    return {t[0]: (t[1], t[2], t[3], t[4], t[5], t[6]) for t in traffic_vols}


if __name__ == '__main__':
    pass